package adt2prism;

import java.io.File;
import java.util.ArrayList;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author zaas
 */
public class ADTreeParser {
    
    public static ADNode ParseTree(String sFile) throws Exception{
        return ADTreeParser.ParseTree(new File(sFile));
    }
    
    private static ADNode ParseTree(File file) throws Exception {
        ADNode tree = null;
        
        //Get the DOM Builder Factory
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

        //Get the DOM Builder
        DocumentBuilder db = dbf.newDocumentBuilder();
        
        //Load and Parse the XML document
 	//doc contains the complete XML in a tree-like structure
        Document doc = db.parse(file);
        doc.getDocumentElement().normalize();
        
        //The root node is <adtree>
        Node root = doc.getDocumentElement();
        
        //Get the children of <adtree>
        ArrayList<Node> children = getXMLNodeChildren(root);
        
        //The tree must have one root, therefore there must be one child of <adtree> 
        if(children.size() != 1)
            throw new Exception("The file contains no tree or the tree is malformed.");
        else{
            // overwrite the root (now root points to the first and only child of <adtree>)
            root = children.get(0);
            
            // build the tree recursively. Assume that the root belongs to the attacker
            tree = buildTree(root, ADConstants.Player.attacker, null);
        }
        return tree;
    }
    
    // Retrieve the children of node n in the XML tree-like representation
    private static ArrayList<Node> getXMLNodeChildren(Node n){
        ArrayList<Node> children = new ArrayList<Node>();
        
        NodeList allChildren = n.getChildNodes();
        for (int i=0; i< allChildren.getLength(); i++){
            if(allChildren.item(i).getNodeType() == Node.ELEMENT_NODE)
                children.add(allChildren.item(i));
        }
        return children;
    }
    
    // Retrieve the value of attribute "refinement" of the XML node. Assume that it is in position 0.
    // To be improved by looking up the name of the attribute.
    private static ADConstants.Refinement getRefinement(Node n){
        String attrib = n.getAttributes().item(0).getTextContent();
        
        ADConstants.Refinement ref = null; 
        switch (attrib) {
            case "disjunctive":
                ref = ADConstants.Refinement.OR;
                break;
            case "conjunctive":
                ref = ADConstants.Refinement.AND;
                break;
            case "seq_conjunctive":
                ref = ADConstants.Refinement.SAND;
                break;
            case "seq_disjunctive":
                ref = ADConstants.Refinement.SOR;
                break;
            case "negative":
                ref = ADConstants.Refinement.NEG;
                break;
        }
        return ref;
    }
    
    // Return true if the value of attribure "switchRole" is yes, false otherwise.
    // Assume that the attribute is in position 1 (second attribute). 
    // To be improved by looking up the name of the attribute.
    private static boolean isSwitchNode(Node n){
        boolean isSwitch = false;

        if (n.getAttributes().item(1) == null)
            isSwitch = false;
        else {
            String attrib = n.getAttributes().item(1).getTextContent();
            if (attrib.equals("yes"))
               isSwitch = true;
        }
        return isSwitch;
    }
    
    //Return true if the node n is a leaf.
    private static boolean isLeaf(Node n){
        boolean isLeaf = true;
        ArrayList<Node> children = getXMLNodeChildren(n);
        
        for(int i=0; i<children.size(); i++){
            if (children.get(i).getNodeName().equals("node")){
                isLeaf = false;
                break;
            }
        }
        return isLeaf;
    }
    
    // Return the cost value of the node n is there exists such an attribute.
    private static double getCost(Node n){
        double cost = Double.NaN;
        ArrayList<Node> children = getXMLNodeChildren(n);
        
        for(int i=0; i<children.size(); i++){  
            if(children.get(i).getNodeName().equals("parameter")){
                NamedNodeMap attribs = children.get(i).getAttributes();
                if(attribs.getNamedItem("name").getNodeValue().equals("cost")){
                    cost = Double.parseDouble(children.get(i).getTextContent());
                    break;
                }
            }
        }
        return cost;
    }
    
    // Return the probability value of the node n is there exists such an attribute.
    private static double getProbability(Node n){
        double prob = Double.NaN;
        ArrayList<Node> children = getXMLNodeChildren(n);
        
        for(int i=0; i<children.size(); i++){
            if(children.get(i).getNodeName().equals("parameter")){
                NamedNodeMap attribs = children.get(i).getAttributes();
                if(attribs.getNamedItem("name").getNodeValue().equals("likelihood")){
                    String sProb = children.get(i).getTextContent();
                    try{
                        prob = Double.parseDouble(sProb);
                        break;
                    } 
                    catch(NumberFormatException e) {
                        System.err.println("Error in parsing probabilities");
                        break;
                    }
                }
            }
        }
        return prob;
    }
    
    private static ADNode buildTree(Node n, ADConstants.Player player, ADNode parent){
        ArrayList<Node> children = getXMLNodeChildren(n);
        ADNode tree = null;
        boolean isLeaf = isLeaf(n);
        
        //if the node n is XML is a leaf in the ADT
        if(isLeaf) {

            if (isSwitchNode(n))
                player = ADConstants.changePlayer(player);

            //get the label of the node in XML, creat the node in the tree with the label
            Node label = children.get(0);
            tree = new ADNode(label.getTextContent(), ADConstants.Refinement.LEAF, player, parent);
            //set the cost and probability values for the leaf in the ADT
            tree.setCost(getCost(n));
            tree.setProbability(getProbability(n));
        }
        else if(!isLeaf){
            Node label = children.get(0);
            
            // get the refinement label of the current node
            ADConstants.Refinement ref = getRefinement(n);
            
            // if the node is a swtichRole node then ignore the refinement in the file and use SIM
            boolean isSwitch = false;
            if (isSwitchNode(n)) {
                //ref = ADConstants.Refinement.SIM;
                isSwitch = true;
                player = ADConstants.changePlayer(player);
            }
            tree = new ADNode(label.getTextContent(), ref, player, parent);
            
            // for each child, add the child to tree. Start from 1 because child 0 is the label node
            for (int i = 1; i < children.size(); i++) {
                    tree.addChild(buildTree(children.get(i),player, tree));
            }
        }
        return tree;
    }
}
